#include "prjlibs-include/standards.h"
#include <stddef.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>

#include "skalibs/include/stddjb.h"
#include "skalibs/include/ttymodes.h"
#include "cvm/include/cvm/client.h"
#include "prjlibs-include/constants.h"
#include "prjlibs-include/diewarn.h"
#include "constants.h"

char const* PROG="cvmlogin";
#define service (PROG+3)

static int read_line(int const fd, char* buf, size_t limit,
                     size_t* const plen) {
  size_t len=0;
  while (limit!=0) {
    ssize_t const nread=read(fd, buf, 1);
    if (nread<0) return -1;
    if (nread==0) break;
    --limit;
    ++len;
    if (*buf=='\n') break;
    ++buf;
  }
  *plen=len;
  return 0;
}

static buffer* logbuf;
static char logpid[(sizeof (pid_t)*CHAR_BIT+2)/3+2]="";

static void logmsg(char const* const x0, char const* const x1) {
  if (logpid[0]=='\0')
    logpid[fmt_ulong(logpid, getpid())]=' ';
  if (buffer_puts(logbuf, service)<0) _exit(100);
  if (buffer_puts(logbuf, ": ")<0) _exit(100);
  if (buffer_puts(logbuf, logpid)<0) _exit(100);
  if (buffer_puts(logbuf, x0)<0) _exit(100);
  if (x1!=NULLP)
    if (buffer_puts(logbuf, x1)<0) _exit(100);
  if (buffer_putsflush(logbuf, "\n")<0) _exit(100);
}

static void check(char* const buf, size_t len, size_t const max,
                  char const* const name) {
  if (len==0 || buf[len-1]!='\n') DIE2X(100, "EOF while reading ", name);
  buf[len-1]='\0';
  if (len==max) DIE2X(100, name, " too long");
  --len;
  if (byte_chr(buf, len, '\0')!=len)
    DIE2X(100, name, " contains null characters");
}

static char const var_cvm[]="$CVM_PLAIN";

int main(int argc, char** argv) {
  stralloc sa=GEN_ALLOC_ZERO;
  char const* x;
  typedef char buf[512];
  static size_t const bufmax=(sizeof (buf))-1;
  buf ubuf, pbuf;
  size_t ulen, plen;
  ttymodes modes, modes_noecho;
  int ret, err;
  char const* const cvm=env_get(var_cvm+1);

  logbuf=buffer_2;

  if (cvm==NULLP) DIE_NOTSET(var_cvm);
  if (ttymodes_gett(STDIN_FILENO, &modes)!=0) DIE0(tcgetattr);
  modes_noecho=modes;
  ttymodes_echo_off(&modes_noecho);

  logmsg("tty ", env_get("TTY"));

  x=env_get("PROMPT_LOGIN");
  if (x==NULLP) x="login: ";
  if (buffer_putsflush(buffer_1, x)<0) DIE1(write, "login prompt");
  if (read_line(STDIN_FILENO, ubuf, bufmax, &ulen)!=0) DIE1(read, "username");
  check(ubuf, ulen, bufmax, "username");

  logmsg("attempt ", ubuf);

  if (env_get("DENY")!=NULLP) {
    if (stralloc_copys(&sa, "ALLOW_")==0 ||
        stralloc_cats(&sa, ubuf)==0 ||
        stralloc_0(&sa)==0)
      DIE0(alloc);
    err=(env_get(sa.s)==NULLP);
  } else {
    if (stralloc_copys(&sa, "DENY_")==0 ||
        stralloc_cats(&sa, ubuf)==0 ||
        stralloc_0(&sa)==0)
      DIE0(alloc);
    err=(env_get(sa.s)!=NULLP);
  }
  if (err) {
    x="login denied";
    logmsg(x, NULLP);
    logbuf=buffer_1;
    logmsg(x, NULLP);
    _exit(100);
  }

  x=env_get("PROMPT_PASSWORD");
  if (x==NULLP) x="Password: ";
  if (buffer_putsflush(buffer_1, x)<0) DIE1(write, "password prompt");
  if (ttymodes_sett(STDIN_FILENO, &modes_noecho)!=0) DIE0(tcsetattr);
  ret=read_line(STDIN_FILENO, pbuf, bufmax, &plen);
  err=errno;
  if (ttymodes_sett(STDIN_FILENO, &modes)!=0) DIE0(tcsetattr);
  if (ret!=0) {
    errno=err;
    DIE1(read, "password");
  }
  if (buffer_putsflush(buffer_1, "\n")<0) DIE1(write, "newline");
  check(pbuf, plen, bufmax, "password");

  {
    char const* credentials[]={ NULLP, NULLP };
    credentials[0]=pbuf;
    ret=cvm_authenticate(cvm, ubuf, NULLP, credentials, 0);
  }
  if (ret!=0) {
    if (ret==CVME_PERMFAIL) {
      x="login incorrect";
      logmsg(x, NULLP);
      logbuf=buffer_1;
      logmsg(x, NULLP);
      _exit(100);
    }
    switch (ret) {
      case CVME_GENERAL:     x="unspecified error";        break;
      case CVME_BAD_CLIDATA: x="invalid data from client"; break;
      case CVME_BAD_MODDATA: x="invalid data from module"; break;
      case CVME_IO:          x="I/O error";                break;
      default:               x="unknown error";            break;
    }
    DIE2X(111, "error authenticating: ", x);
  }

  logmsg("login ok", NULLP);

  {
    pid_t pid=fork();
    switch (pid) {
      case -1: DIE0(fork);
      case 0: break;
      default:
        pid=wait_pid(NULLP, pid);
        err=errno;
        logmsg("logout", NULLP);
        if (pid>=0) _exit(0);
        errno=err;
        DIE0(wait);
    }
  }

  if (pathexec_env(var_user+1, cvm_fact_username)==0) DIE0(alloc);
  if (stralloc_copyb(&sa, "", 0)==0) DIE0(alloc);
  {
    char const* c;
    if (cvm_fact_str(CVM_FACT_USERID, &c)==0)
      if (pathexec_env(VAR_UID+1, c)==0) DIE0(alloc);
    if (cvm_fact_str(CVM_FACT_GROUPID, &c)==0) {
      if (pathexec_env(VAR_GID+1, c)==0) DIE0(alloc);
      if (env_get("INCLUDE_GID")!=NULLP)
        if (stralloc_copys(&sa, " ")==0 ||
            stralloc_cats(&sa, c)==0)
          DIE0(alloc);
    }
    while (cvm_fact_str(CVM_FACT_SUPP_GROUPID, &c)==0)
      if (stralloc_cats(&sa, " ")==0 ||
          stralloc_cats(&sa, c)==0)
        DIE0(alloc);
  }
  x=env_get("EXTRA_GROUPS");
  if ((x!=NULLP && stralloc_cats(&sa, x)==0) ||
      stralloc_0(&sa)==0 ||
      pathexec_env(var_groups+1, sa.s)==0 ||
      pathexec_env(var_name+1, cvm_fact_realname)==0 ||
      pathexec_env(var_home+1, cvm_fact_directory)==0 ||
      pathexec_env(var_shell+1, cvm_fact_shell)==0)
    DIE0(alloc);

  if (dup2(STDOUT_FILENO, STDERR_FILENO)!=STDERR_FILENO) DIE0(dup2);

  {
    char* command[]={ (char*)"setstate", (char*)"lofgGuh", (char*)"loginshell",
                      NULLP };
    if (argc>1) ++argv;
    else argv=command;
    pathexec((char const**)argv);
    DIE1(exec, argv[0]);
  }
}
